/*
   DbltImages.prg

   -- Purpose ----------------------------------------------------------------

   Display images from tables in web pages generated by dBLT

   -- Description ------------------------------------------------------------

   The basic problem of getting images in tables (usually in binary fields) to
   web pages is that there is no good way to include the image data inline
   with the HTML. Instead, IMG tags refer to the image; then -- if the browser
   does not have image loading disabled -- the browser will ask the web
   server (usually but not necessarily the same one that provided the HTML)
   for each image. By this time, the server has long forgotten any association
   with the desired binary fields.

   So the idea here is that while generating the page, and having the data
   module in hand to do so, image data is cached in a separate central table.
   A key that refers to that particular image is returned so that it can be
   included in the IMG element-SRC attribute's URL. When the browser later
   requests that URL, the image data can be returned with the proper HTTP
   headers.

   -- History ----------------------------------------------------------------

      Ken Chan
         v0.01  2001-05-20  Initial 
         v0.02  2001-06-28  Fix locking (lockRetryCount, addEmptyBlock)

   ---------------------------------------------------------------------------
*/ 


//============================================================================
//
// SQL statements to create the cache table and its indexes
//
#define TABLE_NAME "dbltimagecache"
#define CREATE_TABLE_STMT ;
  "create table '" + TABLE_NAME + ".dbf' " + ;
  "(id autoinc," + ;
  " rand integer," + ;
  " last timestamp," + ;
  " image blob(0,1)," + ;
  " primary key (id))"
#define CREATE_INDEX_STMT ;
  "create index last on '" + TABLE_NAME + ".dbf' (last)"

//============================================================================
//
// Constants that control the cache behavior
//
#define MAX_AGE_MINUTES  5   // Number of minutes a newly stored images will not be overwritten
#define EMPTY_BLOCK_RECS 10  // Number of rows to add at once when there are no rows to recycle

//============================================================================
//
// Main routine
//
parameters cIdKey
//set procedure to dbltdata    additive

if empty( cIdKey )
  test()
else
  try 
    new DbltHttpImage().fetch( cIdKey )
  catch ( Exception x )
    quit
  endtry
endif

//close procedure dbltdata


//============================================================================
//
// Test function
//
function test
  //d = new Database()
  //d.databaseName := "MUGS"
  //d.active := true

  q = new Query()
  //q.database := d
  q.sql := "select * from INVENTORY"
  q.active := true

  x = new DbltImageCache()
  c = x.putBinaryField( q.rowset.fields["picture"] )
  ? c
  ? substr(q.rowset.fields["picture"].value,9) == x.getImage( c )


//
// HTTP image fetcher
//
class DbltHttpImage of CgiSession from ":dQuery:dQWebClass.cc"
  this.cache = new DbltImageCache()

  function fetch( cIdKey )
    this.connect()
    local sImage

    try
      if empty( cIdKey )
        cIdKey := this[ "id" ]
      endif
      sImage = this.cache.getImage( cIdKey )
      if not empty( sImage )
        // Lowercase "type" for CGI header
        this.fOut.puts( "Content-type: " + this.imageType( sImage ) )
        // Uppercase "Length" for HTTP header
        this.fOut.puts( "Content-Length: " + len( sImage ) )
        this.fOut.puts( "" )
        this.fOut.write( sImage )
      endif
    catch ( Exception x )
      // Continue to 404
    endtry

    this.fOut.puts( "Status: 404" )
    this.fOut.puts( "" )

  function imageType( sBytes )
    if left( sBytes, 3 ) == "GIF"
      return "image/gif"
    else
      return "image/jpeg"
    endif

endclass

//
// Image cache controller
//
class DbltImageCache
  this.database = new DbltDatabaseConnector().getDatabase()
  this.query    = null

  function getRowset
    if empty( this.query )
      this.query := this.createQuery()
    endif
    return this.query.rowset

  function createQuery
    if not this.database.tableExists( TABLE_NAME )
      this.database.executeSQL( CREATE_TABLE_STMT )
      this.database.executeSQL( CREATE_INDEX_STMT )
    endif
    local q
    this.query = new Query()
    if empty( this.database.databaseName )
      this.query.session := new Session()
      this.query.session.lockRetryCount := 1
    else
      this.query.session  := this.database.session
      this.query.database := this.database
    endif
    this.query.sql := "select * from " + TABLE_NAME
    this.query.active := true
    return this.query

  function putBinaryField( oField )
    return this.putBinaryFieldValue( oField.value )

  function putBinaryFieldValue( sValue )
    if len( sValue ) > 8
      return this.putImage( substr( sValue, 9 ) )
    endif
    return this.putImage( null )
   
  function putImage( sBytes )
    local rand, row
    rand = this.randomKey()
    row = this.emptyRow()
    row.fields[ "rand"  ].value := rand
    row.fields[ "last"  ].value := new Date()
    row.fields[ "image" ].value := sBytes
    row.save()
    row.unlock()
    return this.encodeIdKey( row.fields[ "id" ].value, rand )

  function randomKey()
    return int( rand() * (10^9 - 2) ) + 1

  function encodeIdKey( nId, nRandKey )
    return str( nRandKey, 9, 0, "0" ) + nId

  function decodeKey( cIdKey )
    return val( left( cIdKey, 9 ) )

  function decodeId( cIdKey )
    return val( substr( cIdKey, 10 ) )

  function emptyRow
    local row
    row = this.getRowset()
    row.indexName := "last"
    if row.first()
      local now
      now = new Date()
      do while not row.fields[ "last" ].value + MAX_AGE_MINUTES / 1440 > now
        if row.lockRow()
          return row
        elseif not row.next()
          exit
        endif
      enddo
    endif
    return this.addEmptyBlock( row )

  function addEmptyBlock( oRowset )
    local n
    // Start loop at 2, so there's actually one less blank row added
    // than indicated by the constant
    for n = 2 to EMPTY_BLOCK_RECS
      oRowset.beginAppend()
      oRowset.modified := true
      oRowset.save()
    endfor
    // Create final blank row that will actually be filled in (because
    // it is a row buffer, no lock is needed)
    oRowset.beginAppend()
    return oRowset

  function getImage( cIdKey )
    local nId, nRandKey, row
    nId = this.decodeId( cIdKey )
    nRandKey = this.decodeKey( cIdKey )
    row = this.getRowset()
    row.indexName := "id"
    if row.findKey( nId )
      if row.fields[ "rand" ].value == nRandKey
        // Force type to string instead of memo value
        return "" + row.fields[ "image" ].value
      endif
    endif
    return null

endclass

class DbltDatabaseConnector
  this.alias = "dblt"

  function getDatabase
    try
      return this.aliasedDatabase()
    catch ( Exception dbx )
      return this.defaultDatabase()
    endtry

  function aliasedDatabase()
    local d
    d = new Database()
    d.session = new Session()
    d.session.lockRetryCount := 1
    d.databaseName := this.alias
    d.active := true
    return d

  function defaultDatabase()
    return _app.databases[1]

endclass
  
